/*
 * Copyright 2012  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.tizenopensource.org/license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore.h>
#include <Ecore_X.h>
#include <utilX.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <aul.h>

#include "main.h"
#include "log.h"
#include "scroller.h"
#include "logic.h"
#include "util_efl.h"
#include "clock.h"
#include "pkgmgr.h"

#define PAGE_MOVE 1
#define PAGE_APP_NUM 4
#define PAGE_MAX_NUM 10

#define USE_MAPBUF 0

static struct {
	int menu_pid;
	Eina_Bool is_multi_touched;
	Ecore_Timer *select_item_timer;
	Ecore_Event_Handler *mouse_down_event_handler;
} item_event_info = {
	.menu_pid = 0,
	.is_multi_touched = EINA_FALSE,
	.select_item_timer = NULL,
	.mouse_down_event_handler = NULL,
};

Evas_Object *create_apps_scroller(Evas_Object *parent, void *data);

static void _menu_apps_selected_timer_cb(void *data)
{
	_D("%s : %s", __func__, data);

	item_event_info.menu_pid = 0;
	_D("set item_event_info.menu_pid : 0");

	if(item_event_info.select_item_timer) {
		ecore_timer_del(item_event_info.select_item_timer);
		item_event_info.select_item_timer = NULL;
		_D("delete item_event_info.select_item_timer");
	}
}

static void _more_apps_selected_timer_cb(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	ad->flag_launching_more_apps = EINA_FALSE;
	if(ad->moreapps_timer){
		ecore_timer_del(ad->moreapps_timer);
		ad->moreapps_timer = NULL;
	}
}

void moreapps_click_cb(void *data, Evas_Object *o, const char *emission, const char *source)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	Evas_Object *win_apps = NULL;
	Evas_Object *bg_win_apps = NULL;

	if(ad->flag_launching_more_apps){
		_D("now more apps is launching...");
		return;
	}

	/* create main window */
	if(ad->win_apps != NULL) {
		evas_object_del(ad->win_apps);
		ad->win_apps = NULL;
		_D("More Apps window is aleady exist.");
	}
	ad->flag_launching_more_apps = EINA_TRUE;

	win_apps = _add_window("MORE_APPS", EINA_FALSE);
	retm_if(win_apps == NULL, "Failed add window\n");
	evas_object_resize(win_apps, WIN_SIZE, WIN_SIZE);
	evas_object_move(win_apps, 0, 0);
	ad->win_apps = win_apps;
	_D("create more apps window\n");

	bg_win_apps = _create_bg(win_apps);
	if(bg_win_apps == NULL){
		_E("failed to get parent window");
		return;
	}

	evas_object_smart_callback_add(ad->win_apps, "focus,in", launcher_focus_in_cb, ad);
	evas_object_smart_callback_add(ad->win_apps, "focus,out", launcher_focus_out_cb, ad);

	/* create main layout */
	Evas_Object *ly_apps = NULL;
	ly_apps = _add_layout(ad->win_apps, EDJ_APP, "layout_more_apps");
	retm_if(ly_apps == NULL, "Failed to add more apps layout\n");
	evas_object_size_hint_weight_set(ly_apps, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(ad->win_apps, ly_apps);
	ad->ly_apps = ly_apps;
	evas_object_show(ad->win_apps);

	/* make scroller */
	Evas_Object *scroller_apps = create_apps_scroller(ad->ly_apps, ad);
	if(scroller_apps == NULL) {
		_D("create_apps_scroller() failed");
		evas_object_del(ad->ly_apps);
		ad->flag_launching_more_apps = EINA_FALSE;
		return;
	}
	elm_object_part_content_set(ad->ly_apps, "sw.scroller.apps", scroller_apps);
	ad->scroller_apps = scroller_apps;
	ad->moreapps_timer = ecore_timer_add(0.5, _more_apps_selected_timer_cb, ad);
}

static void _menuapps_item_click_cb(void *data, Evas_Object *o, const char *emission, const char *source)
{
	_D("%s : %s", __func__, data);

	if(item_event_info.is_multi_touched) {
		item_event_info.is_multi_touched = EINA_FALSE;
		_D("multi touched : ignore");
		return;
	}

	item_event_info.is_multi_touched = EINA_FALSE;

	/* check App launching */
	if(item_event_info.menu_pid != 0) {
		_D("App is launching");
		return;
	}

#if USE_AUL_LAUNCH
	bundle *b = NULL;
	//int pid = 0;
	b = bundle_create();
	if(!b) {
		_D("bundle_create() faild");
		return;
	}

	bundle_add(b, "__APP_SVC_OP_TYPE__", SERVICE_OPERATION_MAIN);

	item_event_info.menu_pid = aul_launch_app(data, b);
	if(item_event_info.menu_pid < AUL_R_OK) {
		_D("Launching app ret : [%d]", item_event_info.menu_pid);
		item_event_info.menu_pid = 0;
		return;
	} else {
		_D("add item selected timer");
		item_event_info.select_item_timer = ecore_timer_add(1.0, _menu_apps_selected_timer_cb, data);
	}

	if(b) {
		bundle_free(b);
		_D("bundle free");
	}
#else
	int ret_aul = -1;

	ret_aul = aul_open_app(data);
	if(ret_aul < AUL_R_OK) {
		_D("Launching app ret : [%d]", ret_aul);
		return;
	}
	if(item_event_info.menu_pid < AUL_R_OK) {
		_D("Launching app ret : [%d]", item_event_info.menu_pid);
		item_event_info.menu_pid = 0;
		return;
	} else {
		_D("add item selected timer");
		item_event_info.select_item_timer = ecore_timer_add(1.0, _menu_apps_selected_timer_cb, data);
	}
#endif
}

static void _box_layout_cb(Evas_Object *o, Evas_Object_Box_Data *priv, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	Evas_Coord x=0, y=0, w=0, h=0;
	int size_w = 0;
	int n_children = 0;
	Evas_Object_Box_Option *opt = NULL;
	Evas_Coord child_w=0, child_h=0;
	Eina_List *l = NULL;
	Eina_List *l_next = NULL;

	evas_object_geometry_get(o, &x, &y, &w, &h);

	n_children = eina_list_count(priv->children);
	if(n_children == 0)
		return;

	opt = eina_list_data_get(priv->children);
	if(opt == NULL)
		return;
	if(opt->obj == NULL)
		return;
	evas_object_size_hint_min_get(opt->obj, &child_w, &child_h);

	int order_children = 0;
	EINA_LIST_FOREACH_SAFE(priv->children, l, l_next, opt)
	{
		evas_object_move(opt->obj, x + WIN_SIZE * order_children, y);
		order_children++;
	}

	size_w = WIN_SIZE * order_children;
	evas_object_size_hint_min_set(o, size_w, -1);
	ad->total_page = order_children;
	_D("box total size : %d * %d = %d", WIN_SIZE, order_children, size_w);
}

static int _get_total_scroller_page(void *data)
{
	_D("%s", __func__);
	int page = 0;
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, -1);

	//int page = (ad->app_count - 1)/PAGE_APP_NUM;
	if (ad->more_apps_list->len > 0) {
		 page = (ad->more_apps_list->len- 1)/PAGE_APP_NUM;

		if(page > PAGE_MAX_NUM) {
			return PAGE_MAX_NUM;
		}
	}

	return page;
}

Eina_Bool _mouse_up_cb(void *data, int type, void *event_info)
{
	_D("DBG");
	Ecore_Event_Mouse_Button *button = event_info;
	retv_if(button == NULL, ECORE_CALLBACK_PASS_ON);

	if(button->multi.device != 0) {
		_D("DBG multi touched");
		item_event_info.is_multi_touched = EINA_TRUE;
		return ECORE_CALLBACK_PASS_ON;
	}
	item_event_info.is_multi_touched = EINA_FALSE;
	return ECORE_CALLBACK_PASS_ON;
}

static void _create_apps_contents(Evas_Object *box, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);
	ret_if(box == NULL);

	Evas_Object *layout = NULL;
	Evas_Object *icon[PAGE_APP_NUM];
#if USE_MAPBUF
	Evas_Object *mapbuf = NULL;
#endif

	int page = 0;
	int total_page = 0;
	int num = 0;
	char buf_sw[128] = { 0, };
	char buf_emission[128] = { 0, };
	char buf_txt[128] = { 0, };

	total_page = _get_total_scroller_page(ad);
	more_apps_info *info = NULL;

	item_event_info.mouse_down_event_handler = ecore_event_handler_add(ECORE_EVENT_MOUSE_BUTTON_UP, _mouse_up_cb, ad);
	if(item_event_info.mouse_down_event_handler == NULL) {
		_E("ecore_event_handler_add() failed");
	}

	for(page = 0; page <= total_page; page++) {
		layout = elm_layout_add(box);
		elm_layout_file_set(layout, EDJ_APP, "app_tray");

		for(num = 0; num < PAGE_APP_NUM; num++) {
			if((num+(PAGE_APP_NUM*page)) > ad->more_apps_list->len - 1) {
				_D("This page(%d) has %d icons", page, num);
				break;
			}

#ifdef EMULATOR_TYPE
			if ((num+(PAGE_APP_NUM*page)) < ad->more_apps_list->len) {
				info = g_array_index(ad->more_apps_list, more_apps_info*, (num+(PAGE_APP_NUM*page)));
			}
#else
			info = g_array_index(ad->more_apps_list, more_apps_info*, (num+(PAGE_APP_NUM*page)));
#endif
			if(info == NULL) {
				_D("This page(%d) has %d icons", page, num);
				break;
			}

			snprintf(buf_sw, sizeof(buf_sw), "sw.icon.%d", num);
			snprintf(buf_emission, sizeof(buf_emission), "clicked,rect,icon,%d", num);
			snprintf(buf_txt, sizeof(buf_txt), "txt.app.name.%d", num);

			if(info->icon_path) {
				icon[num] = make_icon(ad, ad->ly_apps, info->icon_path);
				elm_object_part_content_set(layout, buf_sw, icon[num]);
				elm_object_part_text_set(layout, buf_txt, info->app_name);
				elm_object_signal_callback_add(layout, buf_emission, "*", _menuapps_item_click_cb, info->app_id);
				snprintf(buf_emission, sizeof(buf_emission), "clicked,txtrect,icon,%d", num);
				elm_object_signal_callback_add(layout, buf_emission, "*", _menuapps_item_click_cb, info->app_id);
			} else {
				_D("This page(%d) has %d icons", page, num);
				break;
			}

			memset(buf_sw, 0x0, sizeof(buf_sw));
			memset(buf_emission, 0x0, sizeof(buf_emission));
			memset(buf_txt, 0x0, sizeof(buf_txt));
		}

		evas_object_resize(layout, WIN_SIZE, WIN_SIZE);
		evas_object_show(layout);
		ad->page_layout_apps[page] = layout;

#if USE_MAPBUF
		// mapbuf for menu_apps
		mapbuf = elm_mapbuf_add(box);
		elm_mapbuf_enabled_set(mapbuf, EINA_TRUE);
		elm_mapbuf_smooth_set(mapbuf, EINA_TRUE);
		elm_mapbuf_alpha_set(mapbuf, EINA_TRUE);
		evas_object_size_hint_weight_set(mapbuf, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
		evas_object_size_hint_align_set(mapbuf, EVAS_HINT_FILL, EVAS_HINT_FILL);
		elm_object_content_set(mapbuf, layout);
		evas_object_show(mapbuf);
		ad->mapbuf_apps[page] = mapbuf;
		elm_box_pack_end(box, mapbuf);
#else
		elm_box_pack_end(box, layout);
#endif
	}
}

Evas_Object *create_apps_scroller(Evas_Object *parent, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, NULL);
	retv_if(parent == NULL, NULL);

	Evas_Object *scroller = NULL;
	Evas_Object *box = NULL;

	scroller = elm_scroller_add(parent);
	if(scroller == NULL) {
		_E("elm_scroller_add() failed");
		return NULL;
	}
	elm_scroller_loop_set(scroller, EINA_TRUE, EINA_FALSE);
	elm_object_theme_set(scroller, ad->theme);
	elm_object_style_set(scroller, "w-launcher-app/scrollbar");

	evas_object_size_hint_weight_set(scroller, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	evas_object_size_hint_align_set(scroller, EVAS_HINT_FILL, EVAS_HINT_FILL);
	//@TODO: need to check
	//elm_scroller_page_relative_set(scroller, 1.0, 0.0);

	elm_scroller_bounce_set(scroller, EINA_TRUE, EINA_FALSE);
	elm_scroller_policy_set(scroller,ELM_SCROLLER_POLICY_AUTO, ELM_SCROLLER_POLICY_AUTO);
	elm_scroller_page_scroll_limit_set(scroller, PAGE_MOVE, PAGE_MOVE);
	elm_scroller_content_min_limit(scroller, EINA_FALSE, EINA_TRUE);

	elm_object_scroll_lock_y_set(scroller, EINA_TRUE);
	elm_object_tree_focus_allow_set(scroller, EINA_FALSE);
	evas_object_show(scroller);

	elm_scroller_page_size_set(scroller, WIN_SIZE, WIN_SIZE);

	box = elm_box_add(scroller);
	if(box == NULL) {
		_E("elm_box_add() failed");
		evas_object_del(scroller);
		return NULL;
	}
	evas_object_size_hint_weight_set(box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	evas_object_size_hint_fill_set(box, EVAS_HINT_FILL, EVAS_HINT_FILL);
	elm_box_horizontal_set(box, EINA_TRUE);
	evas_object_show(box);
	elm_box_layout_set(box, _box_layout_cb, ad, NULL);
	evas_object_data_set(box, "SCROLLER", scroller);
	_create_apps_contents(box, ad);
	elm_object_content_set(scroller, box);

	ad->menuapps_box = box;

	Eina_List *list = elm_box_children_get(box);
	if(eina_list_count(list) <= 1)
		elm_scroller_loop_set(scroller, EINA_FALSE, EINA_FALSE);

	return scroller;
}

void destroy_apps_scroller(void *data)
{
	struct appdata *ad = data;
	ret_if(ad == NULL);

	if(item_event_info.mouse_down_event_handler) {
		ecore_event_handler_del(item_event_info.mouse_down_event_handler);
		item_event_info.mouse_down_event_handler = NULL;
		_D(" >>> remove ecore event handler");
	}

	if(ad->scroller_apps) {
		evas_object_del(ad->scroller_apps);
		ad->scroller_apps = NULL;
		_D(" >>> remove ad->scroller_apps");
	}
	if(ad->menuapps_box) {
		evas_object_del(ad->menuapps_box);
		ad->menuapps_box = NULL;
		_D(" >>> remove ad->menuapps_box");
	}
	if(ad->ly_apps) {
		evas_object_del(ad->ly_apps);
		ad->ly_apps = NULL;
		_D(" >>> remove ad->ly_apps");
	}
	if(ad->win_apps) {
		evas_object_del(ad->win_apps);
		ad->win_apps = NULL;
		_D(" >>> remove ad->win_apps");
	}
}
