/*
 * Copyright 2012  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.tizenopensource.org/license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore.h>
#include <utilX.h>
#include <dd-display.h>
#include <aul.h>

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include "main.h"
#include "log.h"
#include "logic.h"

#define BUS_NAME "org.tizen.system.deviced"
#define OBJECT_PATH "/Org/Tizen/System/DeviceD"
#define INTERFACE_NAME BUS_NAME

#define DEVICED_PATH_DISPLAY OBJECT_PATH"/Display"
#define DEVICED_INTERFACE_DISPLAY INTERFACE_NAME".display"

#define METHOD_DISPLAY_NAME "PowerKeyIgnore"
#define DBUS_REPLY_TIMEOUT 25000

#define LOW_BATTERY_DBUS_PATH "/Org/Tizen/System/Popup/Lowbat"
#define LOW_BATTERY_DBUS_INTERFACE "org.tizen.system.popup.Lowbat"
#define LOW_BATTERY_DBUS_MEMBER_BATTERY "Extreme"

#define WEB_RUNTIME_STATUS_PATH "/Org/Tizen/Runtime/WRT"
#define WEB_RUNTIME_STATUS_INTERFACE "org.tizen.runtime.wrt.status"
#define WEB_RUNTIME_STATUS_READYDONE "READYDONE"

static int iterfunc(const aul_app_info* info, void* data)
{
	_D("appid : %s, pid %d", info->appid, info->pid);
	if(strcmp(info->appid, APP_NAME_BT_POPUP) == 0) {
		aul_terminate_pid(info->pid);
	}
	return 0;
}

Eina_Bool changed_cb(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, EINA_FALSE);

	ad->flag_wakeup_setting_clock = EINA_TRUE;

	/* Show Idle-Clock */
	if(ad->app_state == APP_STATE_PAUSE) {
		if (ad->win) {
			_D("activate ad->win");
			aul_app_get_running_app_info(iterfunc, NULL);
			elm_win_activate(ad->win);

			if(ad->count_grabed == 0) {
				/* LCD on */
				_D("need_to_lcdon : FALSE");
				ecore_idler_add(display_lcd_on_cb, ad);
			} else {
				_D("need_to_lcdon : TRUE");
				ad->need_to_lcdon = EINA_TRUE;
			}
		}
	} else if(ad->app_state == APP_STATE_RESUME || ad->app_state == APP_STATE_SERVICE) {
		/* LCD on */
		if(ad->count_grabed == 0) {
			/* LCD on */
			_D("need_to_lcdon : FALSE");
			ecore_idler_add(display_lcd_on_cb, ad);
		} else {
			_D("need_to_lcdon : TRUE");
			ad->need_to_lcdon = EINA_TRUE;
		}
	}

	return ECORE_CALLBACK_CANCEL;
}

/* DBus message handler callback */
DBusHandlerResult on_dbus_message_received(DBusConnection *connection, DBusMessage *message, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = data;
	retv_if(ad == NULL, DBUS_HANDLER_RESULT_HANDLED);

	int idx = -1;
	Evas_Coord x, y, w, h;
	int ret_aul;

	if (dbus_message_is_signal(message, WAKEUP_GESTURE_DBUS_INTERFACE, WAKEUP_GESTURE_DBUS_MEMBER_CLOCK))
	{
		// TODO: do the necessary task here!!
		_D("Clock signal received : %p", message);
		if(ad->scroller_apps) {
			evas_object_del(ad->scroller_apps);
			ad->scroller_apps = NULL;
			_D(" >>> remove ad->scroller_apps");
		}
		if(ad->menuapps_box) {
			evas_object_del(ad->menuapps_box);
			ad->menuapps_box = NULL;
			_D(" >>> remove ad->menuapps_box");
		}
		if(ad->ly_apps) {
			evas_object_del(ad->ly_apps);
			ad->ly_apps = NULL;
			_D(" >>> remove ad->ly_apps");
		}
		if(ad->win_apps) {
			evas_object_del(ad->win_apps);
			ad->win_apps = NULL;
			_D(" >>> remove ad->win_apps");
		}

		elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

		if(w != 0) {
			idx = x / w;

			if(idx != ad->idle_clock_page) {
				_D("move idle clock page");
				elm_scroller_page_show(ad->scroller, ad->idle_clock_page, 0);
			}
		}

		ecore_idler_add(changed_cb, ad);
	}
	if (dbus_message_is_signal(message, LOW_BATTERY_DBUS_INTERFACE, LOW_BATTERY_DBUS_MEMBER_BATTERY))
	{
		// TODO: do the necessary task here!!
		_D("Low Battery signal received : %p", message);
		if (ad->win) {
			elm_win_activate(ad->win);
		}

		_D("Clock signal received : %p", message);
		if(ad->scroller_apps) {
			evas_object_del(ad->scroller_apps);
			ad->scroller_apps = NULL;
			_D(" >>> remove ad->scroller_apps");
		}
		if(ad->menuapps_box) {
			evas_object_del(ad->menuapps_box);
			ad->menuapps_box = NULL;
			_D(" >>> remove ad->menuapps_box");
		}
		if(ad->ly_apps) {
			evas_object_del(ad->ly_apps);
			ad->ly_apps = NULL;
			_D(" >>> remove ad->ly_apps");
		}
		if(ad->win_apps) {
			evas_object_del(ad->win_apps);
			ad->win_apps = NULL;
			_D(" >>> remove ad->win_apps");
		}

		elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

		if(w != 0) {
			idx = x / w;

			if(idx != ad->idle_clock_page) {
				_D("move idle clock page");
				elm_scroller_page_show(ad->scroller, ad->idle_clock_page, 0);
			}
		}
	}
	if (dbus_message_is_signal(message, WEB_RUNTIME_STATUS_INTERFACE, WEB_RUNTIME_STATUS_READYDONE))
	{
		_D("receive READYDONE signal from the WRT");
		if(ad->flag_need_to_relaunch){
			_D("need to relaunch web idle clock.");
			ret_aul = aul_open_app(get_app_id(ad->pkgname_idle_clock));
			ad->flag_need_to_relaunch = EINA_FALSE;
			if(ret_aul < AUL_R_OK) {
				_E("Launching app ret : [%d]", ret_aul);
			}
		}
	}

	return DBUS_HANDLER_RESULT_HANDLED;
}

static int append_variant(DBusMessageIter *iter, const char *sig, char *param[])
{
	char *ch;
	int i;
	int int_type;
	uint64_t int64_type;

	if (!sig || !param)
		return 0;

	for (ch = (char*)sig, i = 0; *ch != '\0'; ++i, ++ch) {
		switch (*ch) {
			case 'i':
				int_type = atoi(param[i]);
				dbus_message_iter_append_basic(iter, DBUS_TYPE_INT32, &int_type);
				break;
			case 'u':
				int_type = atoi(param[i]);
				dbus_message_iter_append_basic(iter, DBUS_TYPE_UINT32, &int_type);
				break;
			case 't':
				int64_type = atoi(param[i]);
				dbus_message_iter_append_basic(iter, DBUS_TYPE_UINT64, &int64_type);
				break;
			case 's':
				dbus_message_iter_append_basic(iter, DBUS_TYPE_STRING, &param[i]);
				break;
			default:
				return -EINVAL;
		}
	}

	return 0;
}

DBusMessage *invoke_dbus_method_sync(const char *dest, const char *path,
				const char *interface, const char *method,
						const char *sig, char *param[])
{
	DBusConnection *conn;
	DBusMessage *msg;
	DBusMessageIter iter;
	DBusMessage *reply;
	DBusError err;
	int r;

	conn = dbus_bus_get(DBUS_BUS_SYSTEM, NULL);
	if (!conn) {
		_E("dbus_bus_get error");
		return NULL;
	}

	msg = dbus_message_new_method_call(dest, path, interface, method);
	if (!msg) {
		_D("dbus_message_new_method_call(%s:%s-%s)", path, interface, method);
		return NULL;
	}

	dbus_message_iter_init_append(msg, &iter);
	r = append_variant(&iter, sig, param);
	if (r < 0) {
		_D("append_variant error(%d)", r);
		return NULL;
	}

	dbus_error_init(&err);

	reply = dbus_connection_send_with_reply_and_block(conn, msg, DBUS_REPLY_TIMEOUT, &err);
	if (!reply) {
		_E("dbus_connection_send error(No reply)");
	}

	if (dbus_error_is_set(&err)) {
		_E("dbus_connection_send error(%s:%s)", err.name, err.message);
		reply = NULL;
	}

	dbus_message_unref(msg);
	dbus_error_free(&err);

	return reply;
}

int send_lcd_off_signal(int input)
{
	_D("%s : %d", __func__, input);
	DBusError err;
	DBusMessage *msg;
	int ret, ret_val;
	char *arr[1];
	char val[32];

	snprintf(val, sizeof(val), "%d", input);
	arr[0] = val;

	msg = invoke_dbus_method_sync(BUS_NAME, DEVICED_PATH_DISPLAY, DEVICED_INTERFACE_DISPLAY,
			METHOD_DISPLAY_NAME, "i", arr);

	if (!msg)
		return -EBADMSG;

	dbus_error_init(&err);

	ret = dbus_message_get_args(msg, &err, DBUS_TYPE_INT32, &ret_val, DBUS_TYPE_INVALID);
	if (!ret) {
		_E("no message : [%s:%s]", err.name, err.message);
		ret_val = -EBADMSG;
	}

	dbus_message_unref(msg);
	dbus_error_free(&err);

	_D("%s : %d", DEVICED_INTERFACE_DISPLAY, ret_val);
	return ret_val;
}

static w_launcher_error_e _dbus_sig_attach(char *path, char *interface, char *member)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)w_launcher_app_data_get();
	retv_if(ad == NULL, W_LAUNCHER_ERROR_FAIL);
	retv_if(path == NULL, W_LAUNCHER_ERROR_FAIL);
	retv_if(interface == NULL, W_LAUNCHER_ERROR_FAIL);

	/* DBUS */
	DBusError derror;
	DBusConnection *connection = NULL;

	dbus_error_init(&derror);

	connection = dbus_bus_get_private(DBUS_BUS_SYSTEM, &derror);
	if(!connection) {
		_E("Failed to get dbus : %s", derror.message);
		return W_LAUNCHER_ERROR_FAIL;
	}
	ad->connection = connection;

	dbus_connection_setup_with_g_main(connection, NULL);

	/* Set the DBus rule for the wakeup gesture signal */
	char rules[512];
	sprintf(rules, "path='%s',type='signal',interface='%s',member='%s'", path, interface, member);

	dbus_bus_add_match(connection, rules, &derror);
	if (dbus_error_is_set(&derror)) {
		_E("D-BUS rule adding error: %s", derror.message);
		dbus_error_free(&derror);
		return W_LAUNCHER_ERROR_FAIL;
	}

	/* Set the callback function */
	if(dbus_connection_add_filter(connection, on_dbus_message_received, ad, NULL) == FALSE) {
		_E("Failed to add dbus filter : %s", derror.message);
		dbus_error_free(&derror);
		return W_LAUNCHER_ERROR_FAIL;
	}

	return W_LAUNCHER_ERROR_OK;
}

w_launcher_error_e dbus_gesture_handler_attach(void)
{
	_D("%s", __func__);
	if(W_LAUNCHER_ERROR_OK != _dbus_sig_attach(WAKEUP_GESTURE_DBUS_PATH, WAKEUP_GESTURE_DBUS_INTERFACE, WAKEUP_GESTURE_DBUS_MEMBER_CLOCK))
	{
		_E("Failed to attach gesture dbus");
		return W_LAUNCHER_ERROR_FAIL;
	}

	return W_LAUNCHER_ERROR_OK;
}

w_launcher_error_e dbus_low_battery_handler_attach(void)
{
	_D("%s", __func__);
	if(W_LAUNCHER_ERROR_OK != _dbus_sig_attach(LOW_BATTERY_DBUS_PATH, LOW_BATTERY_DBUS_INTERFACE, LOW_BATTERY_DBUS_MEMBER_BATTERY)) {
		_E("Failed to attach low battery dbus");
		return W_LAUNCHER_ERROR_FAIL;
	}

	return W_LAUNCHER_ERROR_OK;
}

w_launcher_error_e dbus_web_runtime_handler_attach(void)
{
	_D("%s", __func__);
	if(W_LAUNCHER_ERROR_OK != _dbus_sig_attach(WEB_RUNTIME_STATUS_PATH, WEB_RUNTIME_STATUS_INTERFACE, WEB_RUNTIME_STATUS_READYDONE)) {
		_E("Failed to attach low battery dbus");
		return W_LAUNCHER_ERROR_FAIL;
	}

	return W_LAUNCHER_ERROR_OK;
}

static void _dbus_sig_dettach(const char *path, const char *interface, const char *member)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)w_launcher_app_data_get();
	ret_if(ad == NULL);
	ret_if(path == NULL);
	ret_if(interface == NULL);

	DBusError err;

	dbus_error_init(&err);

	/*Disconnecting DBus */
	if(ad->connection) {
		dbus_connection_unref(ad->connection);
		_D("Disconnecting DBUS");
	}

	dbus_connection_remove_filter(ad->connection, on_dbus_message_received, ad);

	char rules[512];
	snprintf(rules, sizeof(rules), "path='%s',type='signal',interface='%s',member='%s'", path, interface, member);

	dbus_bus_remove_match(ad->connection, rules, &err);
	if(dbus_error_is_set(&err)) {
		_E("Failed to dbus_bus_remove_match : %s", err.message);
		dbus_error_free(&err);
	}

	dbus_connection_close(ad->connection);

	return;
}

w_launcher_error_e dbus_gesture_handler_dettach(void)
{
	_D("%s", __func__);
	_dbus_sig_dettach(WAKEUP_GESTURE_DBUS_PATH, WAKEUP_GESTURE_DBUS_INTERFACE, WAKEUP_GESTURE_DBUS_MEMBER_CLOCK);

	return W_LAUNCHER_ERROR_OK;
}

w_launcher_error_e dbus_low_batter_handler_dettach(void)
{
	_D("%s", __func__);
	_dbus_sig_dettach(LOW_BATTERY_DBUS_PATH, LOW_BATTERY_DBUS_INTERFACE, LOW_BATTERY_DBUS_MEMBER_BATTERY);

	return W_LAUNCHER_ERROR_OK;
}

w_launcher_error_e dbus_web_runtime_handler_dettach(void)
{
	_D("%s", __func__);
	_dbus_sig_dettach(WEB_RUNTIME_STATUS_PATH, WEB_RUNTIME_STATUS_INTERFACE, WEB_RUNTIME_STATUS_READYDONE);

	return W_LAUNCHER_ERROR_OK;
}
