/*
 * Copyright 2012  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.tizenopensource.org/license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Ecore.h>
#include <utilX.h>
#include <Ecore_X.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <aul.h>
#include <dd-display.h>
#include <syspopup_caller.h>

#include "main.h"
#include "log.h"
#include "logic.h"
#include "util_efl.h"
#include "scroller.h"
#include "app_tray.h"
#include "clock.h"
#include "pkgmgr.h"
#include "xml_parse.h"
#include "dbus.h"

#define APP_NAME_CAMERA "org.tizen.camera-app"
#define APP_NAME_DIALER "org.tizen.call"
#define APP_NAME_CALL "org.tizen.call"

#define BOOTING_STATE_DONE 3
#define BUFSIZE 1024

#define USE_ROTATION 0
#define APP_NAME_EMULATOR_DEFAULT_CLOCK "org.tizen.idle-clock-digital"
#define SDK_TARGET_BIARY 1

#ifndef EMULATOR_TYPE
#include <context-manager/context_manager.h>
#endif

#define IMGDIR "/usr/apps/org.tizen."PACKAGE"/res/images/w-launcher-app"
#define DEFAULT_ICON_PATH IMGDIR"/b_home_screen_default_icon.png"

int _cache_flush(void *data)
{
	struct appdata *ad = (struct appdata *)data;
	retvm_if(ad == NULL, -1, "Invalid argument: appdata is NULL\n");

	Evas *evas = NULL;

	int file_cache = -1;
	int collection_cache = -1;
	int image_cache = -1;
	int font_cache = -1;

	evas = evas_object_evas_get(ad->win);
	retvm_if(evas == NULL, -1, "Evas is NULL\n");

	file_cache = edje_file_cache_get();
	collection_cache = edje_collection_cache_get();
	image_cache = evas_image_cache_get(evas);
	font_cache = evas_font_cache_get(evas);

	edje_file_cache_set(file_cache);
	edje_collection_cache_set(collection_cache);
	evas_image_cache_set(evas, 0);
	evas_font_cache_set(evas, 0);

	evas_image_cache_flush(evas);
	evas_render_idle_flush(evas);
	evas_font_cache_flush(evas);

	edje_file_cache_flush();
	edje_collection_cache_flush();

	edje_file_cache_set(file_cache);
	edje_collection_cache_set(collection_cache);
	evas_image_cache_set(evas, image_cache);
	evas_font_cache_set(evas, font_cache);

	//appcore_flush_memory();

	return 0;
}

Evas_Object *make_icon(struct appdata *ad, Evas_Object *parent, const char *path)
{
	retvm_if(ad == NULL, NULL, "Invalid argument: appdata is NULL\n");
	_D("icon path : %s", path);
	if(path == NULL) {
		_D("More Apps : set default icon image");
		path = strdup(DEFAULT_ICON_PATH);
	}
	else if(ecore_file_exists(path) != EINA_TRUE) {
		_E("(%s) file is not exist.", path);
		path = strdup(DEFAULT_ICON_PATH);
	}
#if 0
	Evas_Object *icon = elm_image_add(parent);
	if(!elm_image_file_set(icon, path, NULL)) {
		_E("Failed to add icon object");
	}
	elm_image_resizable_set(icon, EINA_TRUE, EINA_TRUE);
	elm_image_orient_set(icon, ELM_IMAGE_ORIENT_NONE);
	//elm_image_aspect_fixed_set(icon, EINA_TRUE);
	elm_image_fill_outside_set(icon, EINA_TRUE);
	evas_object_size_hint_weight_set(icon, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	evas_object_show(icon);
#endif
	Evas_Object *img = evas_object_image_add(evas_object_evas_get(parent));
	evas_object_image_filled_set(img, EINA_TRUE);
	evas_object_image_load_size_set(img, 96, 96);
	evas_object_image_file_set(img, path, NULL);
	evas_object_size_hint_weight_set(img, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	evas_object_resize(img, 128, 128);
	evas_object_show(img);

	return img;
}

static int _launch_app_by_home(void *data)
{
	retv_if(data == NULL, EINA_FALSE);

	_D("%s, data[%s]", __func__, data);

	bundle *b = NULL;
	int pid = 0;
	b = bundle_create();
	if(!b) {
		_E("Failed to create bundle");
		return -1;
	}

	bundle_add(b, "__APP_SVC_OP_TYPE__", SERVICE_OPERATION_MAIN);

	pid = aul_launch_app(data, b);

	if(b) {
		bundle_free(b);
		_D("bundle free");
	}

	return pid;
}

static Eina_Bool _launch_by_power_key(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, EINA_FALSE);

	char *appid = NULL;
	int ret_aul = -1;

	/* set LCD on  - 1 : disable lcd off by power key */
	//send_lcd_off_signal(1);

	appid = vconf_get_str(VCONFKEY_WMS_POWERKEY_DOUBLE_PRESSING);
	if(appid == NULL) {
		_D("Failed to get vconfkey : %s", VCONFKEY_WMS_POWERKEY_DOUBLE_PRESSING);
		appid = strdup(APP_NAME_CAMERA);
	} else {
		appid = strrchr(appid, '/');
		appid++;
	}
	_D("appid : %s", appid);
#if 0
	appid = vconf_get_str(VCONFKEY_SETAPPL_DEFAULT_APP_STR);
	if(appid == NULL) {
		appid = strdup(APP_NAME_CAMERA);
	}
#endif

#if USE_AUL_LAUNCH
	ret_aul = _launch_app_by_home(appid);
#else
	ret_aul = aul_open_app(appid);
#endif
	if(ret_aul < AUL_R_OK) {
		_D("Launching app ret : [%d]", ret_aul);
		free(appid);
		return EINA_FALSE;
	}
	free(appid);
	return EINA_TRUE;
}

static int iterfunc(const aul_app_info* info, void* data)
{
	_D("appid : %s, pid %d", info->appid, info->pid);
	if(strcmp(info->appid, APP_NAME_BT_POPUP) == 0) {
		aul_terminate_pid(info->pid);
	}
	return 0;
}

static Eina_Bool _idler_clock_page_move_cb(void *data)
{
	struct appdata *ad = data;

	if (ad->win) {
		elm_win_activate(ad->win);
		aul_app_get_running_app_info(iterfunc, NULL);
	}
	return EINA_FALSE;
}

static Eina_Bool _powerkey_timer_cb(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, EINA_FALSE);

	Evas_Coord x, y, w, h;
	int idx = -1;

	if(ad->long_press_timer) {
		ecore_timer_del(ad->long_press_timer);
		ad->long_press_timer = NULL;
		_D("delete ad->long_press_timer");
	}

	if(ad->powerkey_timer) {
		ecore_timer_del(ad->powerkey_timer);
		ad->powerkey_timer = NULL;
		_D("delete ad->powerkey_timer");
	}

	if(ad->flag_long_press) {
		_D("power key long poress");
		ad->flag_long_press = EINA_FALSE;
		return ECORE_CALLBACK_CANCEL;
	}

	if(ad->powerkey_count == 2) {
		/* double press */
		_D("powerkey double press");
		ad->powerkey_count = 0;
		if(!_launch_by_power_key(ad)) {
			_E("Failed to launch by double press");
		}
		return ECORE_CALLBACK_CANCEL;
	} else if(ad->powerkey_count == 3) {
		/* triple press */
		_D("powerkey triple press");
		ad->powerkey_count = 0;
		return ECORE_CALLBACK_CANCEL;
	}
	ad->powerkey_count = 0; //initialize powerkey count

	if(!ad->powerkey_lcd_state) {
		_D("just turn LCD on : should not move to idle clock page");
		return ECORE_CALLBACK_CANCEL;
	}

	if (aul_app_is_running(APP_NAME_CALL)) {
		_D("call is running");
		return ECORE_CALLBACK_CANCEL;
	}

	if(ad->scroller_apps) {
		evas_object_del(ad->scroller_apps);
		ad->scroller_apps = NULL;
		_D(" >>> remove ad->scroller_apps");
	}
	if(ad->menuapps_box) {
		evas_object_del(ad->menuapps_box);
		ad->menuapps_box = NULL;
		_D(" >>> remove ad->menuapps_box");
	}
	if(ad->ly_apps) {
		evas_object_del(ad->ly_apps);
		ad->ly_apps = NULL;
		_D(" >>> remove ad->ly_apps");
	}
	if(ad->win_apps) {
		evas_object_del(ad->win_apps);
		ad->win_apps = NULL;
		_D(" >>> remove ad->win_apps");
	}

	elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

	if(w != 0) {
		idx = x / w;

		if(idx != ad->idle_clock_page) {
			elm_scroller_page_show(ad->scroller, ad->idle_clock_page, 0);
			/* set LCD off  - 0 : enable lcd off by power key */
			//send_lcd_off_signal(0);
			//update_time(ad);
			_D("move idle clock page");

			ecore_idler_add(_idler_clock_page_move_cb, ad);
		} else if(idx == ad->idle_clock_page) {
			if(ad->app_state == APP_STATE_PAUSE){
				_D("APP was launched by IDE.");
			}
			else{
				_D("LCD OFF");
				display_change_state(LCD_OFF);
			}

			/* Show Idle-Clock */
			if (ad->win) {
				elm_win_activate(ad->win);
				aul_app_get_running_app_info(iterfunc, NULL);
			}
		}
	}

	return ECORE_CALLBACK_CANCEL;
}

static Eina_Bool _long_press_timer_cb(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	int pid;
	bundle *b;
	retv_if(ad == NULL, EINA_FALSE);

	ad->flag_long_press = EINA_TRUE;
	ad->powerkey_count = 0;

	if(ad->long_press_timer) {
		_D("delete long press timer");
		ecore_timer_del(ad->long_press_timer);
		ad->long_press_timer = NULL;
	}

	b = bundle_create();

	if (!b) return -1;

	pid = syspopup_launch("poweroff-syspopup", b);

	bundle_free(b);

	return ECORE_CALLBACK_CANCEL;
}

static Eina_Bool _key_down_event_cb(void *data, int type, void *event_info)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	int val = -1;
	retv_if(ad == NULL, EINA_FALSE);

	Ecore_Event_Key *ev = event_info;
#if 0
	if(!strcmp(ev->keyname, KEY_POWER)) {
		_D("press POWER Key");

		if(vconf_get_int(VCONFKEY_PM_STATE, &val) < 0) {
			_D("Cannot get VCONFKEY_PM_STATE");
		}

		if(val == VCONFKEY_PM_STATE_NORMAL) {
			_D("LCD state : LCD on");
			ad->powerkey_lcd_state = EINA_TRUE;
		} else if(val == VCONFKEY_PM_STATE_LCDOFF) {
			_D("LCD state : LCD off");
			ad->powerkey_lcd_state = EINA_FALSE;
		}

		ad->powerkey_count++;
		if(ad->powerkey_count > 3) {
			ad->powerkey_count = 1;
		}

		/* long press */
		ad->flag_long_press = EINA_FALSE;

		if(ad->long_press_timer) {
			_D("delete long press timer");
			ecore_timer_del(ad->long_press_timer);
			ad->long_press_timer = NULL;
		}

		_D("create long press timer");
		ad->long_press_timer = ecore_timer_add(0.5, _long_press_timer_cb, ad);
	}
#endif
	return ECORE_CALLBACK_CANCEL;
}

static Eina_Bool _key_up_event_cb(void *data, int type, void *event_info)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, EINA_FALSE);

	Ecore_Event_Key *ev = event_info;
	Evas_Coord x, y, w, h;
	int idx = -1;
	int val = -1;

	if(!strcmp(ev->keyname, KEY_BACK)) {
		if(ad->win_apps) {
			_D("release MENU_APPS BACK Key");
			destroy_apps_scroller(ad);
		} else {
			_D("release MAIN BACK Key");
			elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

			if(w != 0) {
				idx = x / w;

				if(idx != ad->idle_clock_page) {
					elm_scroller_page_show(ad->scroller, ad->idle_clock_page, 0);
					/* set LCD off  - 0 : enable lcd off by power key */
					//send_lcd_off_signal(0);
					//update_time(ad);
					_D("show idle clock");
				} else {
					_D("aul_launch_app : %s", APP_NAME_CAMERA);
					int ret_aul = -1;

#if USE_AUL_LAUNCH
					//ret_aul = _launch_app_by_home(APP_NAME_CAMERA);

					bundle *b = NULL;
					b = bundle_create();
					if(!b) {
						_D("bundle_create() faild");
					}

					bundle_add(b, "launch.effect", "slidedown");
					ret_aul = aul_launch_app(APP_NAME_CAMERA, b);
					if(b) {
						bundle_free(b);
						_D("bundle free");
					}
#else
					ret_aul = aul_open_app(APP_NAME_CAMERA);
#endif
					if(ret_aul < AUL_R_OK) {
						_D("Launching app ret : [%d]", ret_aul);
						return ECORE_CALLBACK_CANCEL;
					}
				}
			}
		}
	} else if(!strcmp(ev->keyname, KEY_MENU)) {
		_D("release MAIN MENU Key");
		elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

		if(w != 0) {
			idx = x / w;

			if(idx == ad->idle_clock_page) {
				_D("aul_launch_app : %s", APP_NAME_DIALER);
				int ret_aul = -1;
#if USE_AUL_LAUNCH
				ret_aul = _launch_app_by_home(APP_NAME_DIALER);
#else
				ret_aul = aul_open_app(APP_NAME_DIALER);
#endif
				if(ret_aul < AUL_R_OK) {
					_D("Launching app ret : [%d]", ret_aul);
					return ECORE_CALLBACK_CANCEL;
				}
			}
		}
	}
#if 0
	else if(!strcmp(ev->keyname, KEY_POWER)) {
		_D("release POWER Key -");

		if (aul_app_is_running(APP_NAME_CALL)) {
			_D("call is running");
			/* set LCD off  - 0 : enable lcd off by power key */
			//send_lcd_off_signal(0);
			display_change_state(LCD_OFF);
		}

		if(ad->long_press_timer) {
			ecore_timer_del(ad->long_press_timer);
			ad->long_press_timer = NULL;
			_D("delete long press timer");
		}

		if(ad->powerkey_timer) {
			_D("delete powerkey timer");
			ecore_timer_del(ad->powerkey_timer);
			ad->powerkey_timer = NULL;
		}

		_D("create powerkey timer");
		ad->powerkey_timer = ecore_timer_add(0.15, _powerkey_timer_cb, ad);
		//_powerkey_timer_cb(ad);
	}
#endif
	return ECORE_CALLBACK_CANCEL;
}

static void _pm_state_changed_cb(keynode_t * node, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	int val = -1;
	int idx = -1;
	Evas_Coord x, y, w, h;

	if(vconf_get_int(VCONFKEY_PM_STATE, &val) < 0) {
		_E("Failed to get VCONFKEY_PM_STATE");
		return;
	}

	if(val == VCONFKEY_PM_STATE_NORMAL) {
		_D("LCD on");
		elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);
		if(w != 0) {
			idx = x / w;
			if(idx == ad->idle_clock_page) {
				/* set LCD off  - 0 : enable lcd off by power key */
				//send_lcd_off_signal(0);
			}
		}
	}

#if 0
	if(val == VCONFKEY_PM_STATE_LCDOFF) {
		_D("LCD off");
		if(ad->timer) {
			pause_time(ad);
		}
	}
#endif
}

static void free_more_apps_list(GArray *list)
{
	_D("%s", __func__);
	int total = list->len;
	int index = 0;

	if (total <= 0) {
		return;
	}

	more_apps_info* info = NULL;

	for (index = 0; index < total; index++) {
		info = g_array_index(list, more_apps_info*, 0);

		if (info->app_id != NULL) {
			free(info->app_id);
		}
		if (info->app_name != NULL) {
			free(info->app_name);
		}
		if (info->icon_path != NULL) {
			free(info->icon_path);
		}
		free(info);

		g_array_remove_index(list, 0);
	}
}

#define DELETE_CMD "delete"
static void _desktop_cb(keynode_t *node, void *data)
{
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	char *event;
	char type[10];
	char package[BUFSIZE];

	char *pkgname = NULL;
	int ret_aul = 0;

	event = vconf_get_str(vconf_keynode_get_name(node));
	ret_if(NULL == event);

	if (sscanf(event, "%10[^:]:%1023s", type, package) != 2) {
		_E("Failed to parse the event format : [%s], [%s]", type, package);
	}

	_D("command[%s], package[%s]", type, package);

	if(!strncmp(type, DELETE_CMD, strlen(DELETE_CMD))){
		if(!strcmp(package, ad->appid_idle_clock)){
			_D("idle clock is deleted. need to launch default idle clock.");
			vconf_set_str(VCONFKEY_WMS_CLOCKS_SET_IDLE, APP_NAME_EMULATOR_DEFAULT_CLOCK);
		}
	}

	/* initialize more_apps_list */
	if (ad->more_apps_list != NULL) {
		int total = ad->more_apps_list->len;
		_D("more_apps_list length : %d", ad->more_apps_list->len);

		if (total > 0) {
			free_more_apps_list(ad->more_apps_list);
		}
	}

	w_launcher_load_app_list(ad);

	if(ad->win_apps) {
		_D("menu_apps is alive");
		/* make scroller */
		elm_object_part_content_unset(ad->ly_apps, "sw.scroller.apps");
		if(ad->menuapps_box) {
			evas_object_del(ad->menuapps_box);
			ad->menuapps_box = NULL;
			_D(" >>> remove ad->menuapps_box");
		}
		if(ad->scroller_apps) {
			evas_object_del(ad->scroller_apps);
			ad->scroller_apps = NULL;
			_D(" >>> remove ad->scroller_apps");
		}

		ad->scroller_apps = (Evas_Object *)create_apps_scroller(ad->ly_apps, ad);
		if(ad->scroller_apps == NULL) {
			_D("Failed to create menu_apps_scroller");
			evas_object_del(ad->ly_apps);
			ad->ly_apps = NULL;
			return;
		}
		elm_object_part_content_set(ad->ly_apps, "sw.scroller.apps", ad->scroller_apps);
		return;
	}

	free(event);
}

static void _bg_color_changed_cb(keynode_t *node, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	int val = -1;

	if(vconf_get_int(VCONFKEY_SETAPPL_HOME_BG_COLOR_TYPE_INT, &val) < 0) {
		_D("Failed to get vconfkey : %s", VCONFKEY_SETAPPL_HOME_BG_COLOR_TYPE_INT);
		return;
	}

	switch(val) {
		case 1:
			_D("RGB : 0, 0, 0");
			elm_object_signal_emit(ad->ly_main, "change,bg,color,1", "bg.color");
			break;
		case 2:
			_D("RGB : 92, 209, 229");
			elm_object_signal_emit(ad->ly_main, "change,bg,color,2", "bg.color");
			break;
		case 3:
			_D("RGB : 255, 102, 0");
			elm_object_signal_emit(ad->ly_main, "change,bg,color,3", "bg.color");
			break;
		case 4:
			_D("RGB : 153, 56, 0");
			elm_object_signal_emit(ad->ly_main, "change,bg,color,4", "bg.color");
			break;
		default:
			_D("default");
			elm_object_signal_emit(ad->ly_main, "change,bg,color,1", "bg.color");
			break;
	}
}

static void _page_show(void *data, Evas * e, Evas_Object * obj, void *event_info)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;

	elm_scroller_page_show(obj, ad->idle_clock_page, 0);
}

static void _favorite_changed_cb(keynode_t *node, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	int ret = -1;

#if 0
	char *xml_data = NULL;

	xml_data = vconf_get_str(VCONFKEY_WMS_FAVORITES_REORDER);
	if(xml_data == NULL) {
		_D("Failed to get VCONFKEY_WMS_FAVORITES_REORDER");
		return;
	}
#endif

	/* parse XML File */
	//@TODO: need to check
	ret = xml_parse(ad);
	if(ret == W_LAUNCHER_ERROR_FAIL) {
		_E("Failed to parse XML file");
		system("cp /usr/apps/org.tizen.w-launcher-app/data/watchlauncher_settings.xml /opt/usr/share/w-launcher/watchlauncher_settings.xml");
		_favorite_changed_cb(NULL, ad);
		return;
	}

	/* re-make main scroller */
	if(ad->scroller) {
		evas_object_del(ad->scroller);
		ad->scroller = NULL;
	}
	Evas_Object *scroller = create_main_scroller(ad->ly_main, ad);
	if(scroller == NULL) {
		_D("create_main_scroller() failed");
		evas_object_del(ad->ly_main);
	}
	elm_object_part_content_set(ad->ly_main, "sw.scroller", scroller);
	ad->scroller = scroller;

	/* set first page : idle clock */
	evas_object_event_callback_add(ad->scroller, EVAS_CALLBACK_RESIZE, _page_show, ad);
}

#if USE_SETTING_IDLE_CLOCK
static Eina_Bool _terminate_idle_clock_cb(void *data)
{
	_D("%s : %d", __func__, data);
	int pid = (int)data;
	aul_terminate_pid(pid);
	return ECORE_CALLBACK_CANCEL;
}

static void _idle_clock_changed_cb(keynode_t *node, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	ret_if(ad == NULL);

	char *appid = NULL;
	char *pkgname = NULL;
	char *pkgtype = NULL;
	int ret_aul = -1;
#if SDK_TARGET_BIARY
	pkgname = vconf_get_str(VCONFKEY_WMS_CLOCKS_SET_IDLE);
	if(pkgname == NULL) {
		_E("Failed to get VCONFKEY_WMS_CLOCKS_SET_IDLE and set basic idle clock");
		pkgname = strdup(APP_NAME_EMULATOR_DEFAULT_CLOCK);
		if(pkgname == NULL) {
			_E("Failed to duplicate string");
			return;
		}
	}
	_D("appid : %s / ad->pkgname_idle_clock : %s", pkgname, ad->pkgname_idle_clock);
	if(strcmp(pkgname, ad->pkgname_idle_clock)) {
		appid = get_app_id(pkgname);
		if(appid == NULL) {
			_E("Failed to get appid. appid is NULL");
			free(pkgname);
			pkgname == NULL;

			pkgname = strdup(APP_NAME_EMULATOR_DEFAULT_CLOCK);
			if(pkgname == NULL) {
				_E("Failed to duplicate string");
				return;
			}

			ad->pkgname_idle_clock = strdup(pkgname);
			if(ad->pkgname_idle_clock == NULL) {
				_E("Failed to duplicate string");
				free(pkgname);
				pkgname = NULL;
				return;
			}

			appid = get_app_id(ad->pkgname_idle_clock);
			if(appid == NULL) {
				_E("Failed to get appid");
				free(pkgname);
				pkgname = NULL;
				return;
			}
		}

		pkgtype = get_app_type(pkgname);
		if(pkgtype == NULL) {
			_E("Failed to get app_type. app_type is NULL");
			free(appid);
			free(pkgname);
			return;
		}

		_D("pkgtype : %s", pkgtype);
		if(!strcmp(pkgtype, APP_TYPE_WIDGET)){
			ret_aul = aul_open_app(appid);
		}
		else{
			ret_aul = _launch_app_by_home(appid);
		}

		if(ret_aul < AUL_R_OK) {
			_D("Launching app ret : [%d]", ret_aul);
			free(appid);
			free(pkgname);
			free(pkgtype);
			return;
		}

		_D("ret_aul : %d", ret_aul);
		_D("ad->pid_idle_clock will be terminate : %d", ad->pid_idle_clock);
		//ecore_idler_add(_terminate_idle_clock_cb, ad->pid_idle_clock);
		_terminate_idle_clock_cb(ad->pid_idle_clock);

		ad->pkgname_idle_clock = strdup(pkgname);
		ad->pid_idle_clock = ret_aul;
		ad->appid_idle_clock = strdup(appid);
		free(pkgtype);
		free(pkgname);
	}
	free(appid);
#endif
}
#endif

Eina_Bool display_lcd_on_cb(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, EINA_FALSE);

	display_change_state(LCD_NORMAL);

	return ECORE_CALLBACK_CANCEL;
}

static w_launcher_error_e _register_vconf(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, W_LAUNCHER_ERROR_OK);

	if(vconf_notify_key_changed(VCONFKEY_PM_STATE, _pm_state_changed_cb, ad) < 0) {
		_E("Failed to register the pm status callback");
	}

	if(vconf_notify_key_changed(VCONFKEY_AIL_INFO_STATE, _desktop_cb, ad) < 0) {
		_E("Failed to register the desktop callback");
	}

	if(vconf_notify_key_changed(VCONFKEY_SETAPPL_HOME_BG_COLOR_TYPE_INT, _bg_color_changed_cb, ad) < 0) {
		_E("Failed to register the bg color callback");
	}
	_bg_color_changed_cb(NULL, ad);

	if(vconf_notify_key_changed(VCONFKEY_WMS_SYNC_LOGS, _favorite_changed_cb, ad) < 0) {
		_E("Failed to register the WMS_SYNC_LOGS callback");
	}

	if(vconf_notify_key_changed(VCONFKEY_WMS_SYNC_CONTACTS, _favorite_changed_cb, ad) < 0) {
		_E("Failed to register the WMS_SYNC_CONTACTS callback");
	}

	if(vconf_notify_key_changed(VCONFKEY_WMS_FAVORITES_REORDER, _favorite_changed_cb, ad) < 0) {
		_E("Failed to register the WMS_FAVORITE_REORDER callback");
	}

#if USE_SETTING_IDLE_CLOCK
	if(vconf_notify_key_changed(VCONFKEY_WMS_CLOCKS_SET_IDLE, _idle_clock_changed_cb, ad) < 0) {
		_E("Failed to register the WMS_CLOCKS_SET_IDLE callback");
		//return W_LAUNCHER_ERROR_FAIL;
	}
	_idle_clock_changed_cb(NULL, ad);
#endif

	return W_LAUNCHER_ERROR_OK;
}

void unregister_vconf(void)
{
	_D("%s", __func__);
	if(vconf_ignore_key_changed(VCONFKEY_PM_STATE, _pm_state_changed_cb) < 0) {
		_E("Failed to ignore the pm status callback");
	}

	if(vconf_ignore_key_changed(VCONFKEY_AIL_INFO_STATE, _desktop_cb) < 0) {
		_E("Failed to ignore the desktop callback");
	}

	if(vconf_ignore_key_changed(VCONFKEY_SETAPPL_HOME_BG_COLOR_TYPE_INT, _bg_color_changed_cb) < 0) {
		_E("Failed to ignore the bg color callback");
	}

	if(vconf_ignore_key_changed(VCONFKEY_WMS_SYNC_LOGS, _favorite_changed_cb) < 0) {
		_E("Failed to ignore the WMS_SYNC_LOGS callback");
	}

	if(vconf_ignore_key_changed(VCONFKEY_WMS_SYNC_CONTACTS, _favorite_changed_cb) < 0) {
		_E("Failed to ignore the WMS_SYNC_CONTACTS callback");
	}

	if(vconf_ignore_key_changed(VCONFKEY_WMS_FAVORITES_REORDER, _favorite_changed_cb) < 0) {
		_E("Failed to ignore the favorite reorder callback");
	}

#if USE_SETTING_IDLE_CLOCK
	if(vconf_ignore_key_changed(VCONFKEY_WMS_CLOCKS_SET_IDLE, _idle_clock_changed_cb) < 0) {
		_E("Failed to ignore the WMS_CLOCKS_SET_IDLE callback");
	}
#endif
}

int grab_key_new(struct appdata *ad, Ecore_X_Window _xwin, int grab_mode)
{
	_D("%s", __func__);
	retvm_if(ad == NULL, -1, "Invalid argument: appdata is NULL\n");

	Ecore_X_Display *disp = NULL;
	Ecore_X_Window xwin = 0;
	int ret = -1;
	ad->count_grabed++;
	_D("ad->count_grabed : %d", ad->count_grabed);

	disp = ecore_x_display_get();
	retvm_if(disp == NULL, -1, "Failed to get display\n");

	if(_xwin == -1)
	{
		/* TOP_POSITION_GRAB */
		xwin = elm_win_xwindow_get(ad->win);
		retvm_if(xwin == 0, -1, "elm_win_xwindow_get() failed\n");
	}
	else
		xwin = _xwin;

	ret = utilx_grab_key(disp, xwin, KEY_POWER, grab_mode);
	retvm_if(ret < 0, -1, "Failed to grab key down\n");
	retvm_if(ret == 1, -1, "Already grab\n");

	return 0;
}

int ungrab_key_new(struct appdata *ad)
{
	Ecore_X_Window xwin = 0;
	Ecore_X_Display *disp = NULL;

	retvm_if(ad == NULL, -1, "Invalid argument: appdata is NULL\n");
	retvm_if(ad->input_win == 0, -1, "Invalid argument: ad->win is NULL\n");

	xwin = elm_win_xwindow_get(ad->win);
	retvm_if(xwin == 0, -1, "Failed to get xwindow\n");

	disp = ecore_x_display_get();
	retvm_if(disp == NULL, -1, "Failed to get display\n");

	utilx_ungrab_key(disp, ad->input_win, KEY_POWER);
	_D("key ungrabed\n");

	return 0;
}

#if USE_ROTATION
void context_update_cb(int error, context_item_e context, char *context_data, void *user_data, int req_id)
{
   //@TODO: need to check
   if(error == CONTEXT_ERROR_NONE) {
        if(context == CONTEXT_ITEM_AUTO_ROTATION) {
             int angle = 0; //rotate angle
             context_context_data_get_int(context_data, CONTEXT_MOTION_ANGLE, &angle);

             if(angle == CONTEXT_AUTO_ROTATION_0) {
                  _D("0 degree");
             } else if(angle == CONTEXT_AUTO_ROTATION_90) {
                  _D("90 degree");
             } else if(angle == CONTEXT_AUTO_ROTATION_180) {
                  _D("180 degree");
             } else if(angle == CONTEXT_AUTO_ROTATION_270) {
                  _D("270 degree");
             } else if(angle == CONTEXT_AUTO_ROTATION_NONE) {
                  _D("invalid degree");
             }
        }
   }
}
#endif

static int _lang_change_cb(void *data)
{
	_D("%s : Language is changed", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, -1);

	int ret = -1;

	/* parse XML File */
	//@TODO: need to check
	ret = xml_parse(ad);
	if(ret == W_LAUNCHER_ERROR_FAIL) {
		_E("Failed to parse XML file");
		system("cp /usr/apps/org.tizen.w-launcher-app/data/watchlauncher_settings.xml /opt/usr/share/w-launcher/watchlauncher_settings.xml");
		_favorite_changed_cb(NULL, ad);
		return -1;
	}

	/* initialize more_apps_list */
	if (ad->more_apps_list != NULL) {
		int total = ad->more_apps_list->len;
		_D("more_apps_list length : %d", ad->more_apps_list->len);

		if (total > 0) {
			free_more_apps_list(ad->more_apps_list);
		}
	}

	w_launcher_load_app_list(ad);

	if(ad->win_apps) {
		_D("menu_apps is alive");
		/* make scroller */
		elm_object_part_content_unset(ad->ly_apps, "sw.scroller.apps");
		if(ad->menuapps_box) {
			evas_object_del(ad->menuapps_box);
			ad->menuapps_box = NULL;
			_D(" >>> remove ad->menuapps_box");
		}
		if(ad->scroller_apps) {
			evas_object_del(ad->scroller_apps);
			ad->scroller_apps = NULL;
			_D(" >>> remove ad->scroller_apps");
		}

		ad->scroller_apps = (Evas_Object *)create_apps_scroller(ad->ly_apps, ad);
		if(ad->scroller_apps == NULL) {
			_E("Failed to create menu_apps_scroller");
			evas_object_del(ad->ly_apps);
			ad->ly_apps = NULL;
		} else {
			elm_object_part_content_set(ad->ly_apps, "sw.scroller.apps", ad->scroller_apps);
		}
	}

	/* re-make main scroller */
	if(ad->scroller) {
		evas_object_del(ad->scroller);
		ad->scroller = NULL;
	}
	Evas_Object *scroller = create_main_scroller(ad->ly_main, ad);
	if(scroller == NULL) {
		_E("Failed to create favorite scroller");
		evas_object_del(ad->ly_main);
	}
	elm_object_part_content_set(ad->ly_main, "sw.scroller", scroller);
	ad->scroller = scroller;

	/* set first page : idle clock */
	evas_object_event_callback_add(ad->scroller, EVAS_CALLBACK_RESIZE, _page_show, ad);

	return EXIT_SUCCESS;
}

void launcher_focus_in_cb(void *data, Evas_Object *obj, void *ei)
{
	_D("%s", __func__);

	struct appdata *ad = data;
	ret_if(ad == NULL);

	int i, ret = -1;

	/* ungrab POWER key */
	for(i = 0; i < ad->count_grabed; i++) {
		ret = ungrab_key_new(ad);
		if(ret != 0) {
			_E("Failed to ungrab POWER key(ad->win)");
		}
	}
	ad->count_grabed = 0;

	if(ad->need_to_lcdon) {
		_D("need_to_lcdon : TRUE");
		//ecore_idler_add(display_lcd_on_cb, ad);
		ecore_timer_add(0.2, display_lcd_on_cb, ad);
		ad->need_to_lcdon = EINA_FALSE;
	}
}

void launcher_focus_out_cb(void *data, Evas_Object *obj, void *ei)
{
	_D("%s", __func__);

	struct appdata *ad = data;
	ret_if(ad == NULL);

	int ret = -1;

	/* grab POWER key */
	if(ad->count_grabed == 0) {
		ret = grab_key_new(ad, ad->input_win, SHARED_GRAB);
		if(ret != 0) {
			_E("Failed to grab POWER key(ad->win)");
		}
	}
}

void _check_aux(Evas_Object *win, void *data)
{
	const Eina_List *l, *ll;

	int id = -1;
	char *hint = NULL;

	struct appdata *ad = data;
	ret_if(ad == NULL);

	l = elm_win_aux_hints_supported_get(win);
	EINA_LIST_FOREACH(l, ll, hint) {
		if (!strncmp(hint, "wm.comp.win.effect.enable", strlen("wm.comp.win.effect.enable"))) {
			id = elm_win_aux_hint_add(win, "wm.comp.win.effect.enable", "0");
			if (id != -1) {
				ad->id_launch_effect = id;
				_D("set no effect : %d", id);
			}
		}
	}
}

int _app_create(struct appdata *ad)
{
	_D("%s", __func__);
	int ret = -1;
	int ret_aul = -1;
	Evas_Object *win = NULL;
	Evas_Object *bg = NULL;
	char *pkgtype = NULL;
	char *appid = NULL;
	char *pkgname = NULL;

	/* create main window */
	if(ad->win == NULL) {
		win = _add_window(PACKAGE, EINA_TRUE);
		retvm_if(win == NULL, -1, "Failed add window\n");
		_check_aux(win, ad);
		evas_object_resize(win, WIN_SIZE, WIN_SIZE);
		evas_object_move(win, 0, 0);
		ad->win = win;
		_D("create window\n");
	}

	/* create BG */

	bg = _create_bg(win);
	if(bg == NULL){
		_E("failed to get parent window");
		return -1;
	}

	/* create input_window */
	ad->input_win = _add_input_window();
	retvm_if(ad->input_win == 0, -1, "Failed to create input window\n");

	/* load favorite apps list */
	ret = xml_parse(ad);
	if(ret != 0) {
		_D("Failed to get XML content");
	}

	evas_object_smart_callback_add(ad->win, "focus,in", launcher_focus_in_cb, ad);
	evas_object_smart_callback_add(ad->win, "focus,out", launcher_focus_out_cb, ad);

	/* create main layout */
	Evas_Object *ly_main = NULL;
	ly_main = _add_layout(ad->win, EDJ_APP, "layout_main");
	retvm_if(ly_main == NULL, -1, "Failed to add main layout\n");
	evas_object_size_hint_weight_set(ly_main, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(ad->win, ly_main);
	ad->ly_main = ly_main;
	evas_object_show(ad->win);

	Elm_Theme *theme = NULL;
	theme = elm_theme_new();
	elm_theme_ref_set(theme, NULL);
	elm_theme_extension_add(theme, EDJDIR"/scrollbar.edj");
	ad->theme = theme;

	/* load more apps list */
	w_launcher_load_app_list(ad);

	pkgname = vconf_get_str(VCONFKEY_WMS_CLOCKS_SET_IDLE);
	if(pkgname == NULL) {
		_E("Failed to get VCONFKEY_WMS_CLOCKS_SET_IDLE and set basic idle clock");
		pkgname = strdup(APP_NAME_EMULATOR_DEFAULT_CLOCK);
		if(pkgname == NULL) {
			_E("Failed to duplicate string");
			return -1;
		}
	}
	else{
		ad->pkgname_idle_clock = strdup(pkgname);
	}
	_D("pkgname : %s / ad->pkgname_idle_clock : %s", pkgname, ad->pkgname_idle_clock);
	appid = get_app_id(ad->pkgname_idle_clock);
	pkgtype = get_app_type(ad->pkgname_idle_clock);
	_D("pkgtype : %s", pkgtype);
	if(pkgtype == NULL) {
		_E("Failed to get app_type. app_type is NULL");
		free(appid);
		free(pkgname);
		appid = NULL;
		pkgname = NULL;

		pkgname = strdup(APP_NAME_EMULATOR_DEFAULT_CLOCK);
		if(pkgname == NULL) {
			_E("Failed to duplicate string");
			return -1;
		}

		ad->pkgname_idle_clock = strdup(pkgname);
		if(ad->pkgname_idle_clock == NULL) {
			_E("Failed to duplicate string");
			free(pkgname);
			return -1;
		}

		appid = get_app_id(ad->pkgname_idle_clock);
		pkgtype = get_app_type(ad->pkgname_idle_clock);
		_D("pkgtype: %s", pkgtype);
		if(pkgtype == NULL) {
			_E("Failed to get default clock");
			free(appid);
			free(pkgtype);
			free(pkgname);
			return -1;
		}
	}
	if(!strcmp(pkgtype, APP_TYPE_WIDGET)){
		ret_aul = aul_open_app(appid);
	}
	else{
		ret_aul = _launch_app_by_home(appid);
	}
	if(ret_aul < AUL_R_OK) {
		_E("Launching app ret : [%d]", ret_aul);
		if(ret_aul == AUL_R_ETIMEOUT){
			_D("web launchpad is not ready.");
			ad->flag_need_to_relaunch = EINA_TRUE;
		}
	} else {
		ad->pid_idle_clock = ret_aul;
		ad->appid_idle_clock = strdup(appid);
		_D("idle clock pid : %d", ret_aul);
	}

	/* make scroller */
	Evas_Object *scroller = create_main_scroller(ad->ly_main, ad);
	if(scroller == NULL) {
		_D("create_main_scroller() failed");
		evas_object_del(ad->ly_main);
		free(appid);
		free(pkgname);
		return -1;
	}
	elm_object_part_content_set(ad->ly_main, "sw.scroller", scroller);
	ad->scroller = scroller;

	/* add key event */
	ad->key_down_handler = ecore_event_handler_add(ECORE_EVENT_KEY_DOWN, (Ecore_Event_Handler_Cb)_key_down_event_cb, ad);
	ad->key_up_handler = ecore_event_handler_add(ECORE_EVENT_KEY_UP, (Ecore_Event_Handler_Cb)_key_up_event_cb, ad);

#if USE_ROTATION
        /* register rotation callback */
        context_item_e context = CONTEXT_ITEM_AUTO_ROTATION;
        context_set_changed_callback(context, NULL, context_update_cb, NULL, &req_id);
#endif

	/* register vconf */
	ret = _register_vconf(ad);
	if(ret != 0) {
		_D("Failed to register vconf");
	}

	/* language change callback */
	if (appcore_set_event_callback(APPCORE_EVENT_LANG_CHANGE, _lang_change_cb, ad) < 0) {
		_E("Cannot register the language change callback.");
	}

	return ret;
}
