/*
 * Copyright 2012  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.tizenopensource.org/license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <app.h>
#include <aul.h>
#include <Ecore.h>
#include <utilX.h>
#include <unistd.h>
#include <errno.h>
#include <dd-display.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#ifndef EMULATOR_TYPE
#include <context-manager/context_manager.h>
#define USE_ROTATION 0 //USE_ROTATIONS 1
#endif

#include "main.h"
#include "log.h"
#include "util_efl.h"
#include "logic.h"
#include "dbus.h"

#define HOME_SERVICE_KEY "home_op"
#define HOME_SERVICE_VALUE_POWERKEY "powerkey"

struct appdata *g_app_data = NULL;

struct appdata *w_launcher_app_data_get(void)
{
	return g_app_data;
}

static int iterfunc(const aul_app_info* info, void* data)
{
	_D("appid : %s, pid %d", info->appid, info->pid);
	if(strcmp(info->appid, APP_NAME_BT_POPUP) == 0) {
		aul_terminate_pid(info->pid);
	}
	return 0;
}

static bool app_create(void *data)
{
	struct appdata *ad = data;
	retvm_if(ad == NULL, false, "Invalid argument: appdata is NULL\n");

	/* initialize variables */
	ad->id_launch_effect = -1;
	ad->need_to_lcdon = EINA_FALSE;
	ad->win_apps = NULL;
	ad->ly_apps = NULL;
	ad->win_w = 0;
	ad->win_h = 0;
	ad->count_grabed = 0;
	ad->scroller = NULL;
	ad->scroller_apps = NULL;
	ad->main_box = NULL;
	ad->menuapps_box = NULL;
	ad->total_page = 0;
	ad->idle_clock_page = 0;
	ad->timer = NULL;
	ad->powerkey_lcd_state = EINA_FALSE;
	ad->powerkey_timer = NULL;
	ad->long_press_timer = NULL;
	ad->key_down_handler = NULL;
	ad->key_up_handler = NULL;
	ad->flag_long_press = EINA_FALSE;
	ad->flag_mouse_down = EINA_FALSE;
	ad->flag_wakeup_setting_clock = EINA_FALSE;
	ad->powerkey_count = 0;
	ad->flag_launching_more_apps = EINA_FALSE;
	ad->moreapps_timer = NULL;
	ad->flag_need_to_relaunch = EINA_FALSE;

	/* Idle Clock */
	ad->pkgname_idle_clock = NULL;
	ad->pid_idle_clock = 0;
	ad->appid_idle_clock = NULL;

	/* DBUS - Gesture */
	if(W_LAUNCHER_ERROR_OK != dbus_gesture_handler_attach()) {
		_E("Failed to attach gesture dbus callback");
	}

	/* DBUS - Low Battery */
	if(W_LAUNCHER_ERROR_OK != dbus_low_battery_handler_attach()) {
		_E("Failed to attach low battery dbus callback");
	}

	/* DBUS - WRT status */
	if(W_LAUNCHER_ERROR_OK != dbus_web_runtime_handler_attach()) {
		_E("Failed to attach web runtime dbus callback");
	}

	if(_app_create(ad) != 0){
		_E("_app_create() if failed\n");
		return false;
	}

	return true;
}

static void app_terminate(void *data)
{
	struct appdata *ad = data;

	ad->app_state = APP_STATE_TERMINATE;

#if USE_ROTATION
	/* unregister rotation callback */
	context_item_e context = CONTEXT_ITEM_AUTO_ROTATION;
	context_unset_changed_callback(context);
#endif

	/*Disconnecting DBus */
	if(ad->connection) {
		dbus_connection_close(ad->connection);
		dbus_connection_unref(ad->connection);
		ad->connection = NULL;
		_D("Disconnecting DBUS");
	}

	if (ad->win) {
		evas_object_del(ad->win);
		ad->win = NULL;
		_D("delete ad->win");
	}

	unregister_vconf();
	ungrab_key_new(ad);
}

static void app_pause(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = data;
	retm_if(ad == NULL, "Invalid argument: appdata is NULL\n");

	ad->app_state = APP_STATE_PAUSE;
	int ret = -1;
	Eina_Bool check_focus = EINA_FALSE;

	/* set LCD on  - 1 : disable lcd off by power key */
	//send_lcd_off_signal(1);
	check_focus = elm_object_focus_get(ad->win);

	_D("focus state : %d", check_focus);

	/* grab POWER key */
	if(check_focus == EINA_FALSE){
		if(ad->count_grabed == 0) {
			ret = grab_key_new(ad, ad->input_win, SHARED_GRAB);
			if(ret != 0) {
				_E("Failed to grab POWER key(ad->win)");
			}
		}
	}
#if 0
	if(ad->timer) {
		pause_time(ad);
	}
#endif
}

static void app_resume(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = data;
	retm_if(ad == NULL, "Invalid argument: appdata is NULL\n");

	int idx = -1;
	int ret = -1;
	int i = 0;
	Evas_Coord x, y, w, h;

	if(ad->id_launch_effect != -1) {
		elm_win_aux_hint_val_set(ad->win, ad->id_launch_effect, "0");
	}

	ad->app_state = APP_STATE_RESUME;

	/* ungrab POWER key */
	for(i = 0; i < ad->count_grabed; i++) {
		ret = ungrab_key_new(ad);
		if(ret != 0) {
			_E("Failed to ungrab POWER key(ad->win)");
		}
	}
	ad->count_grabed = 0;

	if(ad->flag_wakeup_setting_clock) { //wake_up gesture setting : clock
		/* LCD on */
		//ecore_idler_add(display_lcd_on_cb, ad);
		ad->flag_wakeup_setting_clock = EINA_FALSE;
		return;
	}

	elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

	if(w != 0) {
		idx = x / w;
		if(idx == ad->idle_clock_page) {
			/* set LCD off  - 0 : enable lcd off by power key */
			//send_lcd_off_signal(0);

#if 0
			_D("update time : %d", idx);
			update_time(ad);
#endif
		}
	}
}

static void app_service(service_h service, void *data)
{
	_D("%s", __func__);
	struct appdata *ad = data;
	retm_if(ad == NULL, "Invalid argument: appdata is NULL\n");

	/* powerkey bundle */
	char *service_val = NULL;
	int idx = -1;
	Evas_Coord x, y, w, h;

	service_get_extra_data(service, HOME_SERVICE_KEY, &service_val);
	_D("service value : %s", service_val);

	if(service_val != NULL) {
		if(!strncmp(service_val, HOME_SERVICE_VALUE_POWERKEY, strlen(HOME_SERVICE_VALUE_POWERKEY))) {
			_D("powerkey operation");
			if(ad->scroller_apps) {
				evas_object_del(ad->scroller_apps);
				ad->scroller_apps = NULL;
				_D(" >>> remove ad->scroller_apps");
			}
			if(ad->menuapps_box) {
				evas_object_del(ad->menuapps_box);
				ad->menuapps_box = NULL;
				_D(" >>> remove ad->menuapps_box");
			}
			if(ad->ly_apps) {
				evas_object_del(ad->ly_apps);
				ad->ly_apps = NULL;
				_D(" >>> remove ad->ly_apps");
			}
			if(ad->win_apps) {
				evas_object_del(ad->win_apps);
				ad->win_apps = NULL;
				_D(" >>> remove ad->win_apps");
			}
			elm_scroller_region_get(ad->scroller, &x, &y, &w, &h);

			if(w != 0) {
				idx = x / w;
				if(idx == ad->idle_clock_page) {
					/* set LCD off  - 0 : enable lcd off by power key */
					if(ad->app_state == APP_STATE_PAUSE){
						_D("other app was launched by SDK");
					}
					else{
						display_change_state(LCD_OFF);
					}
				}
				else{
					elm_scroller_page_show(ad->scroller, ad->idle_clock_page, 0);
					_D("move idle clock page");
					if (ad->win) {
						elm_win_activate(ad->win);
						aul_app_get_running_app_info(iterfunc, NULL);
					}
				}
			}
		}
	}
	ad->app_state = APP_STATE_SERVICE;
	if (ad->win) {
		elm_win_activate(ad->win);
	}

}

int main(int argc, char *argv[])
{
	struct appdata ad;

	app_event_callback_s event_callback;

	event_callback.create = app_create;
	event_callback.terminate = app_terminate;
	event_callback.pause = app_pause;
	event_callback.resume = app_resume;
	event_callback.service = app_service;
	event_callback.low_memory = NULL;
	event_callback.low_battery = NULL;
	event_callback.device_orientation = NULL;
	event_callback.language_changed = NULL;
	event_callback.region_format_changed = NULL;

	memset(&ad, 0x0, sizeof(struct appdata));

	g_app_data = &ad;
	_D("app_data:%p", g_app_data);

#if 0
	/* @TODO: need to check */
	if(setenv("ELM_THEME", "tizen", 0) < 0) {
		_D("setenv is failed: %s", strerror(errno));
	}
#endif

	if(setenv("BUFMGR_LOCK_TYPE", "none", 0) < 0) {
		_D("setenv is failed: %s", strerror(errno));
	}

#if 0
	if(setenv("ELM_ENGINE", "gl", 1) < 0) {
		_D("setenv is failed: %s", strerror(errno));
	}
#endif

	return app_efl_main(&argc, &argv, &event_callback, &ad);
}
