/*
 *
 * Copyright (c) 2000 - 2012 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


#include <Elementary.h>
#include <package-manager.h>
#include <pkgmgr-info.h>

#include "main.h"
#include "log.h"
#include "logic.h"

#define BUFSIZE 512
#define APP_NAME_CALL "Call"
#define PKG_NAME_SETTING "org.tizen.clocksetting"

char *get_app_pkgname(const char *appid)
{
	int ret = 0;
	char *pkgname = NULL;
	char *re_pkgname = NULL;
	pkgmgrinfo_appinfo_h handle;
	ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;
	ret = pkgmgrinfo_appinfo_get_pkgname(handle, &pkgname);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return NULL;
	}

	/*after call pkgmgrinfo_appinfo_destroy_appinfo, mainappid is destroyed with handle, so must copy it*/
	re_pkgname = strdup(pkgname);
	_D("pkg name: %s - %s", pkgname, re_pkgname);
	pkgmgrinfo_appinfo_destroy_appinfo(handle);

	return re_pkgname;
}

char *get_app_id(const char *pkgid)
{
	int ret = 0;
	char *mainappid = NULL;
	char *re_mainappid = NULL;
	pkgmgrinfo_pkginfo_h handle;
	ret = pkgmgrinfo_pkginfo_get_pkginfo(pkgid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;
	ret = pkgmgrinfo_pkginfo_get_mainappid(handle, &mainappid);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
		return NULL;
	}

	/*after call pkgmgrinfo_appinfo_destroy_appinfo, mainappid is destroyed with handle, so must copy it*/
	re_mainappid = strdup(mainappid);
	_D("label : %s - %s", mainappid, re_mainappid);
	pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return re_mainappid;
}

char *get_app_type(const char *pkgid)
{
	int ret = 0;
	char *apptype = NULL;
	char *re_apptype = NULL;
	pkgmgrinfo_pkginfo_h handle;
	ret = pkgmgrinfo_pkginfo_get_pkginfo(pkgid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;
	ret = pkgmgrinfo_pkginfo_get_type(handle, &apptype);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_pkginfo_destroy_pkginfo(handle);
		return NULL;
	}

	/*after call pkgmgrinfo_appinfo_destroy_appinfo, mainappid is destroyed with handle, so must copy it*/
	re_apptype = strdup(apptype);
	_D("apptype : %s - %s", apptype, re_apptype);
	pkgmgrinfo_pkginfo_destroy_pkginfo(handle);

	return re_apptype;
}

char *get_app_label(const char *appid)
{
	int ret = 0;
	char *label = NULL;
	char *re_label = NULL;
	pkgmgrinfo_appinfo_h handle;
	ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;
	ret = pkgmgrinfo_appinfo_get_label(handle, &label);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return NULL;
	}

	if (label == NULL || strstr(label, "null") || strstr(label, "NULL")) {
		label = strdup(" ");
	}

	/*after call pkgmgrinfo_appinfo_destroy_appinfo, label is destroyed with handle, so must copy it*/
	re_label = strdup(label);
	_D("label : %s - %s", label, re_label);
	pkgmgrinfo_appinfo_destroy_appinfo(handle);

	return re_label;
}

char *get_app_icon(const char *appid)
{
	int ret = 0;
	char *icon = NULL;
	char *re_icon = NULL;
	pkgmgrinfo_appinfo_h handle;
	ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
	if (ret != PMINFO_R_OK)
		return NULL;
	ret = pkgmgrinfo_appinfo_get_icon(handle, &icon);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_destroy_appinfo(handle);
		return NULL;
	}

	if (icon == NULL) {
		icon = strdup("NULL");
	}

	/*after call pkgmgrinfo_appinfo_destroy_appinfo, icon is destroyed with handle, so must copy it*/
	re_icon = strdup(icon);
	_D("icon name: %s - %s", icon, re_icon);
	pkgmgrinfo_appinfo_destroy_appinfo(handle);

	return re_icon;
}

static int _get_app_enabled(const char *appid)
{
        int ret = 0;
        bool enabled;
        pkgmgrinfo_appinfo_h handle;
        ret = pkgmgrinfo_appinfo_get_appinfo(appid, &handle);
        if (ret != PMINFO_R_OK)
                return -1;
        ret = pkgmgrinfo_appinfo_is_enabled(handle, &enabled);
        if (ret != PMINFO_R_OK) {
                pkgmgrinfo_appinfo_destroy_appinfo(handle);
                return -1;
        }
        _D("appid : %s, app enabled: %d", appid, enabled);
        pkgmgrinfo_appinfo_destroy_appinfo(handle);
        return 0;
}

//#ifdef EMULATOR_TYPE
static int _is_preload_pkg(const char *pkgid)
{
        int ret = 0;
        bool preload = 0;
        pkgmgrinfo_appinfo_h handle;

        if (pkgid == NULL) {
        	return preload;
        }
        else if (!strcmp(pkgid, PKG_NAME_SETTING)){
        	_D("pkgid : %s enable setting pkg.", pkgid);
        	return 0;
        }

        ret = pkgmgrinfo_appinfo_get_appinfo(pkgid, &handle);
        if (ret != PMINFO_R_OK)
                return -1;
        ret = pkgmgrinfo_appinfo_is_preload(handle, &preload);
        if (ret != PMINFO_R_OK) {
                pkgmgrinfo_appinfo_destroy_appinfo(handle);
                return -1;
        }
        _D("pkg preload: %d\n", preload);
        pkgmgrinfo_appinfo_destroy_appinfo(handle);
        return preload;
}
//#endif

static int _app_list_cb(pkgmgrinfo_appinfo_h handle, void *user_data)
{
	struct appdata *ad = user_data;
	retv_if(ad == NULL, -1);

	char *appid = NULL;
	char *icon_path = NULL;
	char *app_name = NULL;
	int enabled = 0;
	more_apps_info *info = NULL;

	info = (more_apps_info *)calloc(1, sizeof(*info));
	if(info == NULL) {
		_E("Failed to allocate memory");
		return -1;
	}

	pkgmgrinfo_appinfo_get_appid(handle, &appid);
	//enabled = _get_app_enabled(appid);
//#ifdef EMULATOR_TYPE
	enabled = _is_preload_pkg(appid);
//#endif

	if(!enabled) {
		app_name = get_app_label(appid);
		if(app_name == NULL) {
			_E("Failed to get app label");
			app_name = strdup(" ");
			if(app_name == NULL) {
				_E("Failed to duplicate string : app_name");
				return -1;
			}
		}

		if(strcmp(app_name, APP_NAME_CALL)) {
			icon_path = get_app_icon(appid);
			if(icon_path == NULL) {
				_E("Failed to get app icon");
				icon_path = strdup("NULL");
				if(icon_path == NULL) {
					_E("Failed to duplicate string : icon_path");
					free(app_name);
					return -1;
				}
			}

			info->app_id = strdup(appid);
			if(info->app_id == NULL) {
				_E("Failed to duplicate string : info->app_id");
				free(app_name);
				free(icon_path);
				return -1;
			}
			info->app_name = strdup(app_name);
			if(info->app_name == NULL) {
				_E("Failed to duplicate string : info->app_name");
				free(app_name);
				free(icon_path);
				return -1;
			}
			info->icon_path = strdup(icon_path);
			if(info->icon_path == NULL) {
				_E("Failed to duplicate string : info->icon_path");
				free(app_name);
				free(icon_path);
				return -1;
			}
			g_array_append_val(ad->more_apps_list, info);

			free(icon_path);
		} else {
			_D("Call does not have menu icon");
		}

		free(app_name);
	}

	return 0;
}

gint sort_more_apps_list(gconstpointer _a, gconstpointer _b)
{
	more_apps_info **info_1 = (more_apps_info **)_a;
	more_apps_info **info_2 = (more_apps_info **)_b;

	/* returns less than zero for first arg is less than second arg, *
	 * zero for equal,						 *
	 * greater zero if first arg is greater than second arg 	 */
	return strcmp((*info_1)->app_name, (*info_2)->app_name);
}

static int _add_filter(void *data)
{
	struct appdata *ad = data;
	retv_if(ad == NULL, -1);

	int ret = 0;
	pkgmgrinfo_appinfo_filter_h handle = NULL;
	retv_if(PMINFO_R_OK != pkgmgrinfo_appinfo_filter_create(&handle), -1);

	static GArray *more_apps_list = NULL;

	more_apps_list = g_array_new(FALSE, FALSE, sizeof(more_apps_info*));
	ad->more_apps_list = more_apps_list;
	_D("create more_apps_list");

#if 1
	ret = pkgmgrinfo_appinfo_filter_add_bool(handle, PMINFO_APPINFO_PROP_APP_NODISPLAY, 0);
	if(ret != PMINFO_R_OK) {
		_E("Failed to add fileter");
		pkgmgrinfo_appinfo_filter_destroy(handle);
		return ret;
	}
#else
	ret = pkgmgrinfo_appinfo_filter_add_string(handle, PMINFO_APPINFO_PROP_APP_CATEGORY, "http://tizen.org/category/idle-clock");
	if(ret != PMINFO_R_OK) {
		_E("Failed to add fileter");
		pkgmgrinfo_appinfo_filter_destroy(handle);
		return ret;
	}
#endif

	ret = pkgmgrinfo_appinfo_filter_foreach_appinfo(handle, _app_list_cb, ad);
	if(ret != PMINFO_R_OK) {
		_E("Failed to add fileter");
		pkgmgrinfo_appinfo_filter_destroy(handle);
		return ret;
	}

	pkgmgrinfo_appinfo_filter_destroy(handle);

	/* sorting */
	if(more_apps_list->len > 0) {
		g_array_sort(ad->more_apps_list, sort_more_apps_list);
	}

	return 0;
}

bool w_launcher_load_app_list(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = data;
	retv_if(ad == NULL, EINA_FALSE);

	int ret;
	ret = _add_filter(ad);
	_D("load app list done.");

	return ret;
}
