/*
 * Copyright 2012  Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.tizenopensource.org/license
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <Ecore.h>
#include <utilX.h>
#include <vconf.h>
#include <vconf-keys.h>
#include <aul.h>
#include <errno.h>

#include <libintl.h>
#include <libxml/encoding.h>
#include <libxml/xmlwriter.h>
#include <libxml/xmlreader.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "main.h"
#include "log.h"
#include "logic.h"
#include "util_efl.h"
#include "scroller.h"
#include "clock.h"
#include "pkgmgr.h"

#define MAX_XML_DATA_NUM 30
#define MAX_FAVORITE_APPS_NUM 9
#define PATH_XML "/opt/usr/share/w-launcher/watchlauncher_settings.xml"
#define XML_CONTENTS "<?xml version=\"1.0\" encoding=\"UTF-8\"?> <data> </data>"

gint sort_favorite_apps_list(gconstpointer _a, gconstpointer _b)
{
	favorite_apps_info **info_1 = (favorite_apps_info **)_a;
	favorite_apps_info **info_2 = (favorite_apps_info **)_b;

	/* returns less than zero for first arg is less than second arg, *
	 * zero for equal,						 *
	 * greater zero if first arg is greater than second arg 	 */
	return ((*info_1)->order > (*info_2)->order) ? 1 : -1;
}

static void free_favorite_apps_list(GArray *list)
{
	_D("%s", __func__);
	int total = list->len;
	int index = 0;

	if (total <= 0) {
		return;
	}

	favorite_apps_info* info = NULL;

	for (index = 0; index < total; index++) {
		info = g_array_index(list, favorite_apps_info*, 0);

		if (info->pkgname != NULL) {
			free(info->pkgname);
		}
		if (info->classname != NULL) {
			free(info->classname);
		}
		free(info);

		g_array_remove_index(list, 0);
	}
}

w_launcher_error_e xml_parse(void *data)
{
	_D("%s", __func__);
	struct appdata *ad = (struct appdata *)data;
	retv_if(ad == NULL, W_LAUNCHER_ERROR_OK);

	xmlTextReaderPtr reader;
	int ret;
	const char *name, *value;
	int dep = 0;
	int nodeType = 0;
	char *element = NULL;
	favorite_apps_info *info = NULL;
	static GArray *favorite_apps_list = NULL;

	if (favorite_apps_list == NULL) {
		favorite_apps_list = g_array_new(FALSE, FALSE, sizeof(favorite_apps_info*));
		_D("create favorite_apps_list");
	} else {
		int total = favorite_apps_list->len;
		_D("favorite_apps_list length : %d", favorite_apps_list->len);

		if (total > 0) {
			free_favorite_apps_list(favorite_apps_list);
		}
	}

	reader = xmlReaderForFile(PATH_XML, NULL, 0);
	if(reader != NULL) {
		ret = xmlTextReaderRead(reader);
		while(ret == 1) {
			name = (const char*)xmlTextReaderConstName(reader);
			if (name == NULL) {
				name = "";
			}
			value = (const char*)xmlTextReaderConstValue(reader);
			dep = xmlTextReaderDepth(reader);
			nodeType = xmlTextReaderNodeType(reader);

			if((dep == 1) && (nodeType == 1)) { // <favorite>
				info = (favorite_apps_info *)malloc(sizeof(favorite_apps_info));
				if(info == NULL){
					_E("malloc failed.");
					return W_LAUNCHER_ERROR_FAIL;
				}
				memset(info, '\0', sizeof(favorite_apps_info));
				_D("<%s>", name);
			} else if((dep == 1) && (nodeType == 15)) { // </favorite>
				g_array_append_val(favorite_apps_list, info);
				_D("</%s>", name);
			} else if((dep == 2) && (nodeType == 1)) {
				element = strdup(name);
				if(element == NULL){
					_D("element is NULL");
					return W_LAUNCHER_ERROR_FAIL;
				}
			} else if((dep == 3) && (nodeType == 3)) {
				if (!strcmp("PackageName", element)) {
					info->pkgname = strdup(value);
					_D("%s : %s / %s", name, value, info->pkgname);
				} else if (!strcmp("ClassName", element)) {
					info->classname = strdup(value);
					_D("%s : %s / %s", name, value, info->classname);
				} else if (!strcmp("packageName", element)) {
					info->pkgname = strdup(value);
					_D("%s : %s / %s", name, value, info->pkgname);
				} else if (!strcmp("className", element)) {
					info->classname = strdup(value);
					_D("%s : %s / %s", name, value, info->classname);
				} else if(!strcmp("screen", element)) {
					info->order = atoi(value);
					_D("%s : %s / %d", name, value, info->order);
				}
			}
			ret = xmlTextReaderRead(reader);
		}

		/* sorting */
		if(favorite_apps_list->len > 0) {
			g_array_sort(favorite_apps_list, sort_favorite_apps_list);
		}

		ad->favorite_apps_list = favorite_apps_list;
		xmlFreeTextReader(reader);
		if(ret != 0) {
			_D("Failed to parse %s", PATH_XML);
		}
	} else {
		_D("Unable to open %s", PATH_XML);
		FILE *xml_file = NULL;

		if ((xml_file = fopen(PATH_XML, "w")) == NULL) {
			_D("file is not exist : %s", strerror(errno));
			return W_LAUNCHER_ERROR_FAIL;
		}

		int ret = fwrite(XML_CONTENTS, sizeof(char), strlen(XML_CONTENTS), xml_file);
		if (ret != strlen(XML_CONTENTS)) {
			_D("Fail to write to file");
			fclose(xml_file);
			return W_LAUNCHER_ERROR_FAIL;
		}

		fclose(xml_file);
		xml_parse(ad);
	}

	return W_LAUNCHER_ERROR_OK;
}
